/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 * 
 * Copyright (©) 2005,2006,2007,2008 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 * 
 * Deliantra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

#ifndef UTIL_H__
#define UTIL_H__

#define DEBUG_POISON 0x00 // poison memory before freeing it if != 0
#define DEBUG_SALLOC  0   // add a debug wrapper around all sallocs
#define PREFER_MALLOC 0   // use malloc and not the slice allocator

#if __GNUC__ >= 3
# define is_constant(c)             __builtin_constant_p (c)
# define expect(expr,value)         __builtin_expect ((expr),(value))
# define prefetch(addr,rw,locality) __builtin_prefetch (addr, rw, locality)
# define noinline                   __attribute__((__noinline__))
#else
# define is_constant(c)             0
# define expect(expr,value)         (expr)
# define prefetch(addr,rw,locality)
# define noinline
#endif

#if __GNUC__ < 4 || (__GNUC__ == 4 || __GNUC_MINOR__ < 4)
# define decltype(x) typeof(x)
#endif

// put into ifs if you are very sure that the expression
// is mostly true or mosty false. note that these return
// booleans, not the expression.
#define expect_false(expr) expect ((expr) ? 1 : 0, 0)
#define expect_true(expr)  expect ((expr) ? 1 : 0, 1)

#include <pthread.h>

#include <cstddef>
#include <cmath>
#include <new>
#include <vector>

#include <glib.h>

#include <shstr.h>
#include <traits.h>

#if DEBUG_SALLOC
# define g_slice_alloc0(s) debug_slice_alloc0(s)
# define g_slice_alloc(s) debug_slice_alloc(s)
# define g_slice_free1(s,p) debug_slice_free1(s,p)
void *g_slice_alloc (unsigned long size);
void *g_slice_alloc0 (unsigned long size);
void g_slice_free1 (unsigned long size, void *ptr);
#elif PREFER_MALLOC
# define g_slice_alloc0(s) calloc (1, (s))
# define g_slice_alloc(s) malloc ((s))
# define g_slice_free1(s,p) free ((p))
#endif

// use C0X decltype for auto declarations until ISO C++ sanctifies them (if ever)
#define auto(var,expr) decltype(expr) var = (expr)

// very ugly macro that basically declares and initialises a variable
// that is in scope for the next statement only
// works only for stuff that can be assigned 0 and converts to false
// (note: works great for pointers)
// most ugly macro I ever wrote
#define statementvar(type, name, value) if (type name = 0) { } else if (((name) = (value)), 1)

// in range including end
#define IN_RANGE_INC(val,beg,end) \
  ((unsigned int)(val) - (unsigned int)(beg) <= (unsigned int)(end) - (unsigned int)(beg))

// in range excluding end
#define IN_RANGE_EXC(val,beg,end) \
  ((unsigned int)(val) - (unsigned int)(beg) <  (unsigned int)(end) - (unsigned int)(beg))

void cleanup (const char *cause, bool make_core = false);
void fork_abort (const char *msg);

// rationale for using (U) not (T) is to reduce signed/unsigned issues,
// as a is often a constant while b is the variable. it is still a bug, though.
template<typename T, typename U> static inline T min (T a, U b) { return (U)a < b ? (U)a : b; }
template<typename T, typename U> static inline T max (T a, U b) { return (U)a > b ? (U)a : b; }
template<typename T, typename U, typename V> static inline T clamp (T v, U a, V b) { return v < (T)a ? (T)a : v >(T)b ? (T)b : v; }

template<typename T, typename U> static inline void min_it (T &v, U m) { v = min (v, (T)m); }
template<typename T, typename U> static inline void max_it (T &v, U m) { v = max (v, (T)m); }
template<typename T, typename U, typename V> static inline void clamp_it (T &v, U a, V b) { v = clamp (v, (T)a, (T)b); }

template<typename T, typename U> static inline void swap (T& a, U& b) { T t=a; a=(T)b; b=(U)t; }

template<typename T, typename U, typename V> static inline T min (T a, U b, V c) { return min (a, min (b, c)); }
template<typename T, typename U, typename V> static inline T max (T a, U b, V c) { return max (a, max (b, c)); }

// sign returns -1 or +1
template<typename T>
static inline T sign (T v) { return v < 0 ? -1 : +1; }
// relies on 2c representation
template<>
inline sint8 sign (sint8 v) { return 1 - (sint8 (uint8 (v) >> 7) * 2); }

// sign0 returns -1, 0 or +1
template<typename T>
static inline T sign0 (T v) { return v ? sign (v) : 0; }

// div, with correct rounding (< 0.5 downwards, >=0.5 upwards)
template<typename T> static inline T div    (T val, T div) { return (val + div / 2) / div; }
// div, round-up
template<typename T> static inline T div_ru (T val, T div) { return (val + div - 1) / div; }
// div, round-down
template<typename T> static inline T div_rd (T val, T div) { return (val          ) / div; }

template<typename T>
static inline T
lerp (T val, T min_in, T max_in, T min_out, T max_out)
{
  return min_out + div   <T> ((val - min_in) * (max_out - min_out), max_in  - min_in);
}

// lerp, round-down
template<typename T>
static inline T
lerp_rd (T val, T min_in, T max_in, T min_out, T max_out)
{
  return min_out + div_rd<T> ((val - min_in) * (max_out - min_out), max_in  - min_in);
}

// lerp, round-up
template<typename T>
static inline T
lerp_ru (T val, T min_in, T max_in, T min_out, T max_out)
{
  return min_out + div_ru<T> ((val - min_in) * (max_out - min_out), max_in  - min_in);
}

// lots of stuff taken from FXT

/* Rotate right. This is used in various places for checksumming */
//TODO: that sucks, use a better checksum algo
static inline uint32_t
rotate_right (uint32_t c, uint32_t count = 1)
{
  return (c << (32 - count)) | (c >> count);
}

static inline uint32_t
rotate_left (uint32_t c, uint32_t count = 1)
{
  return (c >> (32 - count)) | (c << count);
}

// Return abs(a-b)
// Both a and b must not have the most significant bit set
static inline uint32_t
upos_abs_diff (uint32_t a, uint32_t b)
{
  long d1 = b - a;
  long d2 = (d1 & (d1 >> 31)) << 1;

  return d1 - d2;               // == (b - d) - (a + d);
}

// Both a and b must not have the most significant bit set
static inline uint32_t
upos_min (uint32_t a, uint32_t b)
{
  int32_t d = b - a;
  d &= d >> 31;
  return a + d;
}

// Both a and b must not have the most significant bit set
static inline uint32_t
upos_max (uint32_t a, uint32_t b)
{
  int32_t d = b - a;
  d &= d >> 31;
  return b - d;
}

// this is much faster than crossfires original algorithm
// on modern cpus
inline int
isqrt (int n)
{
  return (int)sqrtf ((float)n);
}

// this is only twice as fast as naive sqrtf (dx*dy+dy*dy)
#if 0
// and has a max. error of 6 in the range -100..+100.
#else
// and has a max. error of 9 in the range -100..+100.
#endif
inline int 
idistance (int dx, int dy)
{ 
  unsigned int dx_ = abs (dx);
  unsigned int dy_ = abs (dy);

#if 0
  return dx_ > dy_
      ? (dx_ * 61685 + dy_ * 26870) >> 16
      : (dy_ * 61685 + dx_ * 26870) >> 16;
#else
  return dx_ + dy_ - min (dx_, dy_) * 5 / 8;
#endif
}

/*
 * absdir(int): Returns a number between 1 and 8, which represent
 * the "absolute" direction of a number (it actually takes care of
 * "overflow" in previous calculations of a direction).
 */
inline int
absdir (int d)
{
  return ((d - 1) & 7) + 1;
}

extern ssize_t slice_alloc; // statistics

void *salloc_ (int n)            throw (std::bad_alloc);
void *salloc_ (int n, void *src) throw (std::bad_alloc);

// strictly the same as g_slice_alloc, but never returns 0
template<typename T>
inline T *salloc (int n = 1)     throw (std::bad_alloc) { return (T *)salloc_ (n * sizeof (T));              }

// also copies src into the new area, like "memdup"
// if src is 0, clears the memory
template<typename T>
inline T *salloc (int n, T *src) throw (std::bad_alloc) { return (T *)salloc_ (n * sizeof (T), (void *)src); }

// clears the memory
template<typename T>
inline T *salloc0(int n = 1)     throw (std::bad_alloc) { return (T *)salloc_ (n * sizeof (T), 0);           }

// for symmetry
template<typename T>
inline void sfree (T *ptr, int n = 1) throw ()
{
  if (expect_true (ptr))
    {
      slice_alloc -= n * sizeof (T);
      if (DEBUG_POISON) memset (ptr, DEBUG_POISON, n * sizeof (T));
      g_slice_free1 (n * sizeof (T), (void *)ptr);
      assert (slice_alloc >= 0);//D
    }
}

// nulls the pointer
template<typename T>
inline void sfree0 (T *&ptr, int n = 1) throw ()
{
  sfree<T> (ptr, n);
  ptr = 0;
}

// makes dynamically allocated objects zero-initialised
struct zero_initialised
{
  void *operator new (size_t s, void *p)
  {
    memset (p, 0, s);
    return p;
  }

  void *operator new (size_t s)
  {
    return salloc0<char> (s);
  }

  void *operator new[] (size_t s)
  {
    return salloc0<char> (s);
  }

  void operator delete (void *p, size_t s)
  {
    sfree ((char *)p, s);
  }

  void operator delete[] (void *p, size_t s)
  {
    sfree ((char *)p, s);
  }
};

// makes dynamically allocated objects zero-initialised
struct slice_allocated
{
  void *operator new (size_t s, void *p)
  {
    return p;
  }

  void *operator new (size_t s)
  {
    return salloc<char> (s);
  }

  void *operator new[] (size_t s)
  {
    return salloc<char> (s);
  }

  void operator delete (void *p, size_t s)
  {
    sfree ((char *)p, s);
  }

  void operator delete[] (void *p, size_t s)
  {
    sfree ((char *)p, s);
  }
};

// a STL-compatible allocator that uses g_slice
// boy, this is verbose
template<typename Tp>
struct slice_allocator
{
  typedef size_t size_type;
  typedef ptrdiff_t difference_type;
  typedef Tp *pointer;
  typedef const Tp *const_pointer;
  typedef Tp &reference;
  typedef const Tp &const_reference;
  typedef Tp value_type;

  template <class U> 
  struct rebind
  {
    typedef slice_allocator<U> other;
  };

  slice_allocator () throw () { }
  slice_allocator (const slice_allocator &) throw () { }
  template<typename Tp2>
  slice_allocator (const slice_allocator<Tp2> &) throw () { }

  ~slice_allocator () { }

  pointer address (reference x) const { return &x; }
  const_pointer address (const_reference x) const { return &x; }

  pointer allocate (size_type n, const_pointer = 0)
  {
    return salloc<Tp> (n);
  }

  void deallocate (pointer p, size_type n)
  {
    sfree<Tp> (p, n);
  }

  size_type max_size () const throw ()
  {
    return size_t (-1) / sizeof (Tp);
  }

  void construct (pointer p, const Tp &val)
  {
    ::new (p) Tp (val);
  }

  void destroy (pointer p)
  {
    p->~Tp ();
  }
};

// P. L'Ecuyer, “Maximally Equidistributed Combined Tausworthe Generators”, Mathematics of Computation, 65, 213 (1996), 203–213.
// http://www.iro.umontreal.ca/~lecuyer/myftp/papers/tausme.ps
// http://www.iro.umontreal.ca/~lecuyer/myftp/papers/tausme2.ps
struct tausworthe_random_generator
{
  uint32_t state [4];

  void operator =(const tausworthe_random_generator &src)
  {
    state [0] = src.state [0];
    state [1] = src.state [1];
    state [2] = src.state [2];
    state [3] = src.state [3];
  }

  void seed (uint32_t seed);
  uint32_t next ();
};

// Xorshift RNGs, George Marsaglia
// http://www.jstatsoft.org/v08/i14/paper
// this one is about 40% faster than the tausworthe one above (i.e. not much),
// despite the inlining, and has the issue of only creating 2**32-1 numbers.
// see also http://www.iro.umontreal.ca/~lecuyer/myftp/papers/xorshift.pdf
struct xorshift_random_generator
{
  uint32_t x, y;

  void operator =(const xorshift_random_generator &src)
  {
    x = src.x;
    y = src.y;
  }

  void seed (uint32_t seed)
  {
    x = seed;
    y = seed * 69069U;
  }

  uint32_t next ()
  {
    uint32_t t = x ^ (x << 10);
    x = y;
    y = y ^ (y >> 13) ^ t ^ (t >> 10);
    return y;
  }
};

template<class generator>
struct random_number_generator : generator
{
  // uniform distribution, 0 .. max (0, num - 1)
  uint32_t operator ()(uint32_t num)
  {
    return !is_constant (num) ? get_range (num)                        // non-constant
         : num & (num - 1)    ? (this->next () * (uint64_t)num) >> 32U // constant, non-power-of-two
         :                      this->next () & (num - 1);             // constant, power-of-two
  }

  // return a number within (min .. max)
  int operator () (int r_min, int r_max)
  {
    return is_constant (r_min) && is_constant (r_max) && r_min <= r_max
              ? r_min + operator ()(r_max - r_min + 1)
              : get_range (r_min, r_max);
  }

  double operator ()()
  {
    return this->next () / (double)0xFFFFFFFFU;
  }

protected:
  uint32_t get_range (uint32_t r_max);
  int get_range (int r_min, int r_max);
};

typedef random_number_generator<tausworthe_random_generator> rand_gen;

extern rand_gen rndm, rmg_rndm;

INTERFACE_CLASS (attachable)
struct refcnt_base
{
  typedef int refcnt_t;
  mutable refcnt_t ACC (RW, refcnt);

  MTH void refcnt_inc () const { ++refcnt; }
  MTH void refcnt_dec () const { --refcnt; }

  refcnt_base () : refcnt (0) { }
};

// to avoid branches with more advanced compilers
extern refcnt_base::refcnt_t refcnt_dummy;

template<class T>
struct refptr
{
  // p if not null
  refcnt_base::refcnt_t *refcnt_ref () { return p ? &p->refcnt : &refcnt_dummy; }

  void refcnt_dec ()
  {
    if (!is_constant (p))
      --*refcnt_ref ();
    else if (p)
      --p->refcnt;
  }

  void refcnt_inc ()
  {
    if (!is_constant (p))
      ++*refcnt_ref ();
    else if (p)
      ++p->refcnt;
  }

  T *p;

  refptr () : p(0) { }
  refptr (const refptr<T> &p) : p(p.p) { refcnt_inc (); }
  refptr (T *p) : p(p) { refcnt_inc (); }
  ~refptr () { refcnt_dec (); }

  const refptr<T> &operator =(T *o)
  {
    // if decrementing ever destroys we need to reverse the order here
    refcnt_dec ();
    p = o;
    refcnt_inc ();
    return *this;
  }

  const refptr<T> &operator =(const refptr<T> &o)
  {
    *this = o.p;
    return *this;
  }

  T &operator * () const { return *p; }
  T *operator ->() const { return  p; }

  operator T *() const { return p; }
};

typedef refptr<maptile>   maptile_ptr;
typedef refptr<object>    object_ptr;
typedef refptr<archetype> arch_ptr;
typedef refptr<client>    client_ptr;
typedef refptr<player>    player_ptr;

struct str_hash
{
  std::size_t operator ()(const char *s) const
  {
#if 0
    uint32_t hash = 0;

    /* use the one-at-a-time hash function, which supposedly is
     * better than the djb2-like one used by perl5.005, but
     * certainly is better then the bug used here before.
     * see http://burtleburtle.net/bob/hash/doobs.html
     */
    while (*s)
      {
        hash += *s++;
        hash += hash << 10;
        hash ^= hash >>  6;
      }

    hash += hash <<  3;
    hash ^= hash >> 11;
    hash += hash << 15;
#else
    // use FNV-1a hash (http://isthe.com/chongo/tech/comp/fnv/)
    // it is about twice as fast as the one-at-a-time one,
    // with good distribution.
    // FNV-1a is faster on many cpus because the multiplication
    // runs concurrent with the looping logic.
    uint32_t hash = 2166136261;
    
    while (*s)
      hash = (hash ^ *s++) * 16777619;
#endif

    return hash;
  }
};

struct str_equal
{
  bool operator ()(const char *a, const char *b) const
  {
    return !strcmp (a, b);
  }
};

// Mostly the same as std::vector, but insert/erase can reorder
// the elements, making append(=insert)/remove O(1) instead of O(n).
//
// NOTE: only some forms of erase are available
template<class T>
struct unordered_vector : std::vector<T, slice_allocator<T> >
{
  typedef typename unordered_vector::iterator iterator;

  void erase (unsigned int pos)
  {
    if (pos < this->size () - 1)
      (*this)[pos] = (*this)[this->size () - 1];

    this->pop_back ();
  }

  void erase (iterator i)
  {
    erase ((unsigned int )(i - this->begin ()));
  }
};

// This container blends advantages of linked lists
// (efficiency) with vectors (random access) by
// by using an unordered vector and storing the vector
// index inside the object.
//
// + memory-efficient on most 64 bit archs
// + O(1) insert/remove
// + free unique (but varying) id for inserted objects
// + cache-friendly iteration
// - only works for pointers to structs
//
// NOTE: only some forms of erase/insert are available
typedef int object_vector_index;

template<class T, object_vector_index T::*indexmember>
struct object_vector : std::vector<T *, slice_allocator<T *> >
{
  typedef typename object_vector::iterator iterator;

  bool contains (const T *obj) const
  {
    return obj->*indexmember;
  }

  iterator find (const T *obj)
  {
    return obj->*indexmember
      ? this->begin () + obj->*indexmember - 1
      : this->end ();
  }

  void push_back (T *obj)
  {
    std::vector<T *, slice_allocator<T *> >::push_back (obj);
    obj->*indexmember = this->size ();
  }

  void insert (T *obj)
  {
    push_back (obj);
  }

  void insert (T &obj)
  {
    insert (&obj);
  }

  void erase (T *obj)
  {
    unsigned int pos = obj->*indexmember;
    obj->*indexmember = 0;

    if (pos < this->size ())
      {
        (*this)[pos - 1] = (*this)[this->size () - 1];
        (*this)[pos - 1]->*indexmember = pos;
      }

    this->pop_back ();
  }

  void erase (T &obj)
  {
    erase (&obj);
  }
};

// basically does what strncpy should do, but appends "..." to strings exceeding length
// returns the number of bytes actually used (including \0)
int assign (char *dst, const char *src, int maxsize);

// type-safe version of assign
template<int N>
inline int assign (char (&dst)[N], const char *src)
{
  return assign ((char *)&dst, src, N);
}

typedef double tstamp;

// return current time as timestamp
tstamp now ();

int similar_direction (int a, int b);

// like sprintf, but returns a "static" buffer
const char *format (const char *format, ...);

/////////////////////////////////////////////////////////////////////////////
// threads, very very thin wrappers around pthreads

struct thread
{
  pthread_t id;

  void start (void *(*start_routine)(void *), void *arg = 0);

  void cancel ()
  {
    pthread_cancel (id);
  }

  void *join ()
  {
    void *ret;

    if (pthread_join (id, &ret))
      cleanup ("pthread_join failed", 1);

    return ret;
  }
};

// note that mutexes are not classes
typedef pthread_mutex_t smutex;

#if __linux && defined (PTHREAD_ADAPTIVE_MUTEX_INITIALIZER_NP)
 #define SMUTEX_INITIALISER PTHREAD_ADAPTIVE_MUTEX_INITIALIZER_NP
#else
 #define SMUTEX_INITIALISER PTHREAD_MUTEX_INITIALIZER
#endif

#define SMUTEX(name) smutex name = SMUTEX_INITIALISER
#define SMUTEX_LOCK(name)   pthread_mutex_lock   (&(name))
#define SMUTEX_UNLOCK(name) pthread_mutex_unlock (&(name))

typedef pthread_cond_t scond;

#define SCOND(name) scond name = PTHREAD_COND_INITIALIZER
#define SCOND_SIGNAL(name)     pthread_cond_signal    (&(name))
#define SCOND_BROADCAST(name)  pthread_cond_broadcast (&(name))
#define SCOND_WAIT(name,mutex) pthread_cond_wait      (&(name), &(mutex))

#endif

